/* S/SL table (1280 entries) */

int sslTable[] = {
/*    0 */	oCall,    11, oCall,   165, oCall,   178, oCall,   351, 
/*    8 */	oGenerateDefinitions, oGenerateTable, oReturn, oInputChoice, 
/*   12 */	   34, oInputChoice,    22, oSetParameter, cInputOutput, 
/*   17 */	oSetClass, oCall,    67, oJump,    32,     1,    22,    15, 
/*   25 */	oSetParameter, cInput, oSetClass, oCall,    67, oCall,    40, 
/*   32 */	oJump,    39,     1,    21,    13, oCall,    40, oReturn, 
/*   40 */	oInputChoice,    49, oSetParameter, cOutput, oSetClass, 
/*   45 */	oCall,    67, oJump,    52,     1,    22,    42, 
/*   52 */	oInputChoice,    63, oInput,    22, oSetParameter, 
/*   57 */	cInputOutput, oSetClass, oCall,    67, oJump,    66,     1, 
/*   64 */	   21,    54, oReturn, oInput,     3, oInput,    28, 
/*   71 */	oInputChoice,   114, oCall,  1268, oEnterNewSymbol, 
/*   76 */	oInputChoice,   100, oChooseClass, oChoice,    83, oJump, 
/*   82 */	   92,     3,     1,    81,     2,    81,     3,    81, 
/*   90 */	oError, eIllegalStringSynonym, oCall,  1268, oEnterNewSymbol, 
/*   95 */	oCall,   126, oEnterSecondNewSymbolValue, oJump,   105, 
/*  100 */	    1,     1,    78, oCall,   126, oInput,    28, 
/*  107 */	oIncrementClassValue, oJump,   123, oJump,   125, oJump, 
/*  113 */	  123,     2,     0,    73,     4,   110, oInput,     0, 
/*  121 */	oJump,    73, oJump,    71, oReturn, oInputChoice,   160, 
/*  128 */	oInputChoice,   147, oSetClassValue, oJump,   158, 
/*  133 */	oLookupSymbol, oChooseSymbolClass, oChoice,   141, oError, 
/*  138 */	eUndefinedSymbol, oJump,   144,     1,     0,   137, 
/*  144 */	owSetClassValueFromSymbolValue, oJump,   158,     3,     2, 
/*  149 */	  130,     0,   133,     1,   133, oInput,     2, oJump, 
/*  157 */	  130, oJump,   163,     1,     5,   128, 
/*  163 */	oxEnterNewSymbolValue, oReturn, oInputChoice,   174, 
/*  167 */	oSetParameter, cError, oSetClass, oCall,    67, oJump,   177, 
/*  174 */	    1,    23,   167, oReturn, oInputChoice,   208, oInput, 
/*  181 */	    0, oCall,  1268, oSetParameter, cType, oSetClass, 
/*  187 */	oEnterNewSymbol, oxEnterNewSymbolValue, oIncrementClassValue, 
/*  190 */	oxSetClassFromSymbolValue, oCall,    67, oJump,   215, 
/*  195 */	oInput,     0, oCall,  1268, oSetParameter, cMechanism, 
/*  201 */	oSetClass, oEnterNewSymbol, oSaveEnclosingSymbol, oCall, 
/*  205 */	  218, oJump,   215,     2,    24,   180,    25,   195, 
/*  213 */	oJump,   217, oJump,   178, oReturn, oInput,     3, oInput, 
/*  221 */	   28, oInputChoice,   339, oCall,  1268, 
/*  226 */	oRestoreEnclosingSymbol, oSetParameter, cUpdateOp, oSetClass, 
/*  230 */	oEnterNewSymbol, oxEnterNewSymbolValue, oSaveCurrentSymbol, 
/*  233 */	oInputChoice,   271, oSetParameter, cParmUpdateOp, oSetClass, 
/*  238 */	oChangeSymbolClass, oInput,     0, oLookupSymbol, 
/*  242 */	oChooseSymbolClass, oChoice,   255, 
/*  245 */	oxSetClassFromSymbolValue, oJump,   265, oError, 
/*  249 */	eUndefinedSymbol, oSetParameter, cNotFound, oSetClass, oJump, 
/*  254 */	  265,     2,     5,   245,     0,   248, oError, 
/*  261 */	eIllegalParameterClass, oSetParameter, cNotFound, oSetClass, 
/*  265 */	oRestoreCurrentSymbol, oxEnterSymbolParameterClass, oInput, 
/*  268 */	   13, oJump,   274,     1,    12,   235, oInputChoice, 
/*  275 */	  326, oChooseSymbolClass, oChoice,   289, oSetParameter, 
/*  280 */	cChoiceOp, oSetClass, oJump,   295, oSetParameter, 
/*  285 */	cParmChoiceOp, oSetClass, oJump,   295,     2,     7,   279, 
/*  292 */	    8,   284, oChoiceEnd, oChangeSymbolClass, oInput,     0, 
/*  298 */	oLookupSymbol, oChooseSymbolClass, oChoice,   312, 
/*  302 */	oxSetClassFromSymbolValue, oJump,   322, oError, 
/*  306 */	eUndefinedSymbol, oSetParameter, cNotFound, oSetClass, oJump, 
/*  311 */	  322,     2,     5,   302,     0,   305, oError, 
/*  318 */	eIllegalResultClass, oSetParameter, cNotFound, oSetClass, 
/*  322 */	oRestoreCurrentSymbol, oyEnterSymbolResultClass, oJump, 
/*  325 */	  329,     1,    11,   276, oSetClassFromSymbolClass, 
/*  330 */	oIncrementClassValue, oInput,    28, oJump,   348, oJump, 
/*  336 */	  350, oJump,   348,     2,     0,   224,     4,   335, 
/*  344 */	oInput,     0, oJump,   224, oJump,   222, oReturn, oInput, 
/*  352 */	   26, oInput,    28, oStartRules, oInputChoice,   504, 
/*  358 */	oLookupSymbol, oInputChoice,   436, oSetParameter, 
/*  362 */	cChoiceRule, oSetClass, oChooseSymbolClass, oChoice,   382, 
/*  367 */	oxChooseSymbolValue, oChoice,   372, oJump,   377,     1, 
/*  373 */	-9999,   370, oError, eSymbolPreviouslyDefined, oJump,   390, 
/*  379 */	oEnterNewSymbol, oJump,   390,     2,    12,   367,     0, 
/*  386 */	  379, oError, eSymbolPreviouslyDefined, oEnterNewSymbol, 
/*  390 */	oySetClassFromSymbolResultClass, 
/*  391 */	oEnterSymbolValueFromAddress, oSaveCurrentSymbol, oInput, 
/*  394 */	    0, oLookupSymbol, oChooseSymbolClass, oChoice,   426, 
/*  399 */	oChooseClass, oChoice,   407, oxSetClassFromSymbolValue, 
/*  403 */	oRestoreCurrentSymbol, oyEnterSymbolResultClass, oJump, 
/*  406 */	  420,     1,     0,   402, oxVerifySymbolClassValue, 
/*  411 */	oChoice,   415, oJump,   420,     1,     1,   413, oError, 
/*  419 */	eWrongDeclaredResultClass, oJump,   433, oError, 
/*  423 */	eUndefinedSymbol, oJump,   433,     2,     5,   399,     0, 
/*  430 */	  422, oError, eIllegalResultClass, oRestoreCurrentSymbol, 
/*  434 */	oJump,   469,     1,    11,   361, oSetParameter, cRule, 
/*  441 */	oSetClass, oChooseSymbolClass, oChoice,   460, 
/*  445 */	oxChooseSymbolValue, oChoice,   450, oJump,   455,     1, 
/*  451 */	-9999,   448, oError, eSymbolPreviouslyDefined, oJump,   468, 
/*  457 */	oEnterNewSymbol, oJump,   468,     2,    11,   445,     0, 
/*  464 */	  457, oError, eSymbolPreviouslyDefined, oEnterNewSymbol, 
/*  468 */	oEnterSymbolValueFromAddress, oSaveEnclosingSymbol, oInput, 
/*  471 */	    3, oInput,    28, oBeginRule, oCall,   518, oInput, 
/*  478 */	    4, oRestoreEnclosingSymbol, oChooseSymbolClass, oChoice, 
/*  482 */	  491, oEmit, aReturn, oJump,   497, oEmit, aRuleEnd, oJump, 
/*  490 */	  497,     2,    11,   483,    12,   487, oChoiceEnd, 
/*  497 */	oSaveRule, oJump,   513, oJump,   515, oJump,   513,     2, 
/*  505 */	    0,   358,    27,   500, oInput,     0, oJump,   358, 
/*  513 */	oJump,   356, oEndRules, oResolveCalls, oReturn, 
/*  518 */	oInputChoice,   600, oCall,   628, oJump,   623, oCall, 
/*  525 */	  651, oJump,   623, oInputChoice,   534, oJump,   543, 
/*  532 */	oJump,   543,     2,     0,   530,     1,   532, oInput, 
/*  540 */	    0, oJump,   530, oCall,   654, oJump,   623, oInput, 
/*  548 */	    0, oCall,   677, oJump,   623, oInput,     0, oCall, 
/*  556 */	  698, oJump,   623, oCall,   723, oJump,   623, oCall, 
/*  564 */	  791, oJump,   623, oCall,   881, oJump,   623, oCall, 
/*  572 */	  904, oJump,   623, oInputChoice,   589, oCall,   919, 
/*  579 */	oJump,   598, oCall,  1016, oJump,   598, oCall,  1055, 
/*  587 */	oJump,   598,     3,     0,   577,     9,   581,    20, 
/*  595 */	  585, oCall,  1066, oJump,   623,    10,     1,   520, 
/*  603 */	    6,   524,     7,   528,     8,   547,     9,   553, 
/*  611 */	   11,   559,     0,   563,    14,   567,    10,   571, 
/*  619 */	   16,   575, oJump,   627, oInput,    28, oJump,   518, 
/*  627 */	oReturn, oEmit, aInput, oLookupSymbol, oChooseSymbolClass, 
/*  632 */	oChoice,   640, oJump,   649, oError, eUndefinedSymbol, 
/*  638 */	oJump,   649,     3,     1,   634,     3,   634,     0, 
/*  646 */	  636, oError, eWrongSymbolClass, oEmitSymbolValue, oReturn, 
/*  651 */	oEmit, aInputAny, oReturn, oEmit, aEmit, oLookupSymbol, 
/*  657 */	oChooseSymbolClass, oChoice,   666, oJump,   675, oError, 
/*  663 */	eUndefinedSymbol, oJump,   675,     3,     2,   660,     3, 
/*  670 */	  660,     0,   662, oError, eWrongSymbolClass, 
/*  675 */	oEmitSymbolValue, oReturn, oEmit, aError, oLookupSymbol, 
/*  680 */	oChooseSymbolClass, oChoice,   689, oJump,   696, oError, 
/*  686 */	eUndefinedSymbol, oJump,   696,     2,     4,   683,     0, 
/*  693 */	  685, oError, eWrongSymbolClass, oEmitSymbolValue, oReturn, 
/*  698 */	oEmit, aCall, oEmitNullCallAddress, oLookupSymbol, 
/*  702 */	oChooseSymbolClass, oChoice,   715, oEnterCall, oJump,   722, 
/*  708 */	oSetParameter, cRule, oSetClass, oEnterNewSymbol, oEnterCall, 
/*  713 */	oJump,   722,     2,    11,   705,     0,   708, oError, 
/*  721 */	eWrongSymbolClass, oReturn, oRestoreEnclosingSymbol, 
/*  724 */	oInputChoice,   770, oChooseSymbolClass, oChoice,   762, 
/*  729 */	oySetClassFromSymbolResultClass, oLookupSymbol, 
/*  731 */	oChooseSymbolClass, oChoice,   738, oError, eUndefinedSymbol, 
/*  736 */	oJump,   751,     1,     0,   734, oVerifySymbolClass, 
/*  742 */	oChoice,   746, oJump,   751,     1,     1,   744, oError, 
/*  750 */	eWrongResultClass, oEmit, aSetResult, oEmitSymbolValue, 
/*  754 */	oEmit, aReturn, oJump,   768, oError, eIllegalValuedReturn, 
/*  760 */	oJump,   768,     2,    12,   729,    11,   758, oChoiceEnd, 
/*  768 */	oJump,   790,     1,     0,   726, oChooseSymbolClass, 
/*  774 */	oChoice,   784, oEmit, aReturn, oJump,   790, oError, 
/*  781 */	eIllegalNonvaluedReturn, oJump,   790,     2,    11,   776, 
/*  787 */	   12,   780, oChoiceEnd, oReturn, oLookupSymbol, 
/*  792 */	oChooseSymbolClass, oChoice,   799, oCall,   628, oJump, 
/*  798 */	  880,     2,     1,   795,     3,   795, oInputChoice, 
/*  805 */	  860, oChooseSymbolClass, oChoice,   847, 
/*  809 */	ozSetClassFromSymbolParameterClass, oSaveCurrentSymbol, 
/*  811 */	oInput,     0, oLookupSymbol, oChooseSymbolClass, oChoice, 
/*  816 */	  821, oError, eUndefinedSymbol, oJump,   834,     1,     0, 
/*  823 */	  817, oVerifySymbolClass, oChoice,   829, oJump,   834, 
/*  829 */	    1,     1,   827, oError, eWrongParameterClass, oEmit, 
/*  835 */	aSetParameter, oEmitSymbolValue, oRestoreCurrentSymbol, 
/*  838 */	oEmitSymbolValue, oJump,   856, oError, eUndefinedSymbol, 
/*  843 */	oInput,     0, oJump,   856,     2,     8,   809,     0, 
/*  851 */	  841, oError, eWrongSymbolClass, oInput,     0, oInput, 
/*  857 */	   13, oJump,   880,     1,    12,   806, oChooseSymbolClass, 
/*  864 */	oChoice,   873, oEmitSymbolValue, oJump,   880, oError, 
/*  870 */	eUndefinedSymbol, oJump,   880,     2,     7,   866,     0, 
/*  877 */	  869, oError, eWrongSymbolClass, oReturn, oPushCycle, 
/*  882 */	oInput,    28, oCall,   518, oInput,    15, oEmit, aJump, 
/*  890 */	oEmitCycleAddress, oxChooseCycleExits, oChoice,   898, 
/*  894 */	oError, eCycleHasNoExits, oJump,   901,     1,     0,   894, 
/*  901 */	oResolveCycleExits, oPopCycle, oReturn, oChooseCycleDepth, 
/*  905 */	oChoice,   911, oError, eExitNotInCycle, oJump,   918,     1, 
/*  912 */	    0,   907, oEmit, aJump, oEmitNullAddress, 
/*  917 */	oEnterCycleExit, oReturn, oLookupSymbol, oInputChoice,   983, 
/*  922 */	oChooseSymbolClass, oChoice,   967, 
/*  925 */	ozSetClassFromSymbolParameterClass, oSaveCurrentSymbol, 
/*  927 */	oInput,     0, oLookupSymbol, oChooseSymbolClass, oChoice, 
/*  932 */	  937, oError, eUndefinedSymbol, oJump,   950,     1,     0, 
/*  939 */	  933, oVerifySymbolClass, oChoice,   945, oJump,   950, 
/*  945 */	    1,     1,   943, oError, eWrongParameterClass, oEmit, 
/*  951 */	aSetParameter, oEmitSymbolValue, oRestoreCurrentSymbol, 
/*  954 */	oySetClassFromSymbolResultClass, oEmitSymbolValue, oJump, 
/*  957 */	  979, oError, eUndefinedSymbol, oSetParameter, cNotFound, 
/*  962 */	oSetClass, oInput,     0, oJump,   979,     2,    10,   925, 
/*  970 */	    0,   958, oError, eWrongSymbolClass, oSetParameter, 
/*  975 */	cNotFound, oSetClass, oInput,     0, oInput,    13, oJump, 
/*  982 */	 1010,     1,    12,   922, oChooseSymbolClass, oChoice, 
/*  988 */	 1000, oEmitSymbolValue, oySetClassFromSymbolResultClass, 
/*  991 */	oJump,  1010, oError, eUndefinedSymbol, oSetParameter, 
/*  996 */	cNotFound, oSetClass, oJump,  1010,     2,     9,   989, 
/* 1003 */	    0,   993, oError, eWrongSymbolClass, oSetParameter, 
/* 1008 */	cNotFound, oSetClass, oEmit, aChoice, oEmitNullAddress, 
/* 1013 */	oCall,  1075, oReturn, oEmit, aCall, oEmitNullCallAddress, 
/* 1019 */	oInput,     0, oLookupSymbol, oChooseSymbolClass, oChoice, 
/* 1024 */	 1039, oEnterCall, oySetClassFromSymbolResultClass, oJump, 
/* 1028 */	 1049, oSetParameter, cChoiceRule, oSetClass, 
/* 1032 */	oEnterNewSymbol, oEnterCall, oSetParameter, cNotFound, 
/* 1036 */	oSetClass, oJump,  1049,     2,    12,  1025,     0,  1029, 
/* 1044 */	oError, eWrongSymbolClass, oSetParameter, cNotFound, 
/* 1048 */	oSetClass, oEmit, aChoice, oEmitNullAddress, oCall,  1075, 
/* 1054 */	oReturn, oEmit, aSetParameterFromInput, oEmit, aChoice, 
/* 1059 */	oEmitNullAddress, oSetParameter, cInput, oSetClass, oCall, 
/* 1064 */	 1075, oReturn, oEmit, aInputChoice, oEmitNullAddress, 
/* 1069 */	oSetParameter, cInput, oSetClass, oCall,  1075, oReturn, 
/* 1075 */	oInput,    28, oPushChoice, oSaveCurrentSymbol, oInput, 
/* 1080 */	   19, oCall,  1174, oInput,     3, oInput,    28, 
/* 1087 */	oRestoreCurrentSymbol, oChooseSymbolClass, oChoice,  1102, 
/* 1091 */	oyChooseSymbolResultClass, oChoice,  1097, 
/* 1094 */	oyEnterSymbolResultClass, oJump,  1100,     1,     0,  1094, 
/* 1100 */	oJump,  1105,     1,    12,  1091, oCall,   518, 
/* 1107 */	oInputChoice,  1140, oEmit, aJump, oEmitNullAddress, 
/* 1112 */	oxEnterChoiceMerge, oInputChoice,  1127, oInput,     3, 
/* 1117 */	oInput,    28, oxResolveChoiceTableAddress, oEmitChoiceTable, 
/* 1121 */	oCall,   518, oJump,  1169, oJump,  1138,     1,    20, 
/* 1129 */	 1115, oCall,  1174, oInput,     3, oInput,    28, oCall, 
/* 1137 */	  518, oJump,  1167,     1,    19,  1109, oEmit, aJump, 
/* 1145 */	oEmitNullAddress, oxEnterChoiceMerge, 
/* 1147 */	oxResolveChoiceTableAddress, oEmitChoiceTable, 
/* 1149 */	oChooseChoiceClass, oChoice,  1160, oEmit, aInput, 
/* 1154 */	oEmitFirstChoiceValue, oEmit, aJump, 
/* 1157 */	oxEmitFirstChoiceAddress, oJump,  1165,     1,     1,  1152, 
/* 1163 */	oEmit, aChoiceEnd, oJump,  1169, oJump,  1107, oInput,    17, 
/* 1171 */	oResolveChoiceMerges, oPopChoice, oReturn, 
/* 1174 */	ovSetClassFromChoiceClass, oInputChoice,  1256, 
/* 1177 */	oLookupSymbol, oChooseChoiceClass, oChoice,  1185, 
/* 1181 */	oSetClassFromSymbolClass, oChangeChoiceClass, oJump,  1188, 
/* 1185 */	    1,     0,  1181, oChooseClass, oChoice,  1211, 
/* 1191 */	oChooseSymbolClass, oChoice,  1200, oJump,  1209, oError, 
/* 1197 */	eUndefinedSymbol, oJump,  1209,     3,     1,  1194,     3, 
/* 1204 */	 1194,     0,  1196, oError, eWrongLabelClass, oJump,  1234, 
/* 1211 */	    1,     1,  1191, oChooseSymbolClass, oChoice,  1221, 
/* 1217 */	oError, eUndefinedSymbol, oJump,  1234,     1,     0,  1217, 
/* 1224 */	oVerifySymbolClass, oChoice,  1229, oJump,  1234,     1, 
/* 1230 */	    1,  1227, oError, eWrongLabelClass, 
/* 1234 */	oVerifyChoiceSymbolLabel, oChoice,  1239, oJump,  1244, 
/* 1239 */	    1,     1,  1237, oError, eDuplicateLabel, 
/* 1244 */	oEnterChoiceSymbolLabel, oInputChoice,  1249, oJump,  1254, 
/* 1249 */	    1,    18,  1247, oJump,  1267, oJump,  1265,     2, 
/* 1257 */	    0,  1177,     1,  1177, oInput,     0, oJump,  1177, 
/* 1265 */	oJump,  1175, oReturn, oLookupSymbol, oChooseSymbolClass, 
/* 1270 */	oChoice,  1274, oJump,  1279,     1,     0,  1272, oError, 
/* 1278 */	eSymbolPreviouslyDefined, oReturn, 0
};
