/*
 * Note: this file originally auto-generated by mib2c using
 *  : mib2c.column_enums.conf,v 5.2 2003/02/22 04:09:25 hardaker Exp $
 */
#ifndef MTA_ENUMS_H
#define MTA_ENUMS_H

#endif                          /* MTA_ENUMS_H */
