/*
 * Note: this file originally auto-generated by mib2c using
 *  : mib2c.column_defines.conf,v 5.1 2002/05/08 05:42:47 hardaker Exp $
 */
#ifndef MTA_COLUMNS_H
#define MTA_COLUMNS_H

/*
 * column number definitions for table mtaGroupTable 
 */
#define COLUMN_MTAGROUPINDEX		1
#define COLUMN_MTAGROUPRECEIVEDMESSAGES		2
#define COLUMN_MTAGROUPREJECTEDMESSAGES		3
#define COLUMN_MTAGROUPSTOREDMESSAGES		4
#define COLUMN_MTAGROUPTRANSMITTEDMESSAGES		5
#define COLUMN_MTAGROUPRECEIVEDVOLUME		6
#define COLUMN_MTAGROUPSTOREDVOLUME		7
#define COLUMN_MTAGROUPTRANSMITTEDVOLUME		8
#define COLUMN_MTAGROUPRECEIVEDRECIPIENTS		9
#define COLUMN_MTAGROUPSTOREDRECIPIENTS		10
#define COLUMN_MTAGROUPTRANSMITTEDRECIPIENTS		11
#define COLUMN_MTAGROUPOLDESTMESSAGESTORED		12
#define COLUMN_MTAGROUPINBOUNDASSOCIATIONS		13
#define COLUMN_MTAGROUPOUTBOUNDASSOCIATIONS		14
#define COLUMN_MTAGROUPACCUMULATEDINBOUNDASSOCIATIONS		15
#define COLUMN_MTAGROUPACCUMULATEDOUTBOUNDASSOCIATIONS		16
#define COLUMN_MTAGROUPLASTINBOUNDACTIVITY		17
#define COLUMN_MTAGROUPLASTOUTBOUNDACTIVITY		18
#define COLUMN_MTAGROUPREJECTEDINBOUNDASSOCIATIONS		19
#define COLUMN_MTAGROUPFAILEDOUTBOUNDASSOCIATIONS		20
#define COLUMN_MTAGROUPINBOUNDREJECTIONREASON		21
#define COLUMN_MTAGROUPOUTBOUNDCONNECTFAILUREREASON		22
#define COLUMN_MTAGROUPSCHEDULEDRETRY		23
#define COLUMN_MTAGROUPMAILPROTOCOL		24
#define COLUMN_MTAGROUPNAME		25
#define COLUMN_MTAGROUPSUCCESSFULCONVERTEDMESSAGES		26
#define COLUMN_MTAGROUPFAILEDCONVERTEDMESSAGES		27
#define COLUMN_MTAGROUPDESCRIPTION		28
#define COLUMN_MTAGROUPURL		29
#define COLUMN_MTAGROUPCREATIONTIME		30
#define COLUMN_MTAGROUPHIERARCHY		31
#define COLUMN_MTAGROUPOLDESTMESSAGEID		32
#define COLUMN_MTAGROUPLOOPSDETECTED		33
#define COLUMN_MTAGROUPLASTOUTBOUNDASSOCIATIONATTEMPT		34

/*
 * column number definitions for table mtaGroupAssociationTable 
 */
#define COLUMN_MTAGROUPASSOCIATIONINDEX		1

/*
 * column number definitions for table mtaTable 
 */
#define COLUMN_MTARECEIVEDMESSAGES		1
#define COLUMN_MTASTOREDMESSAGES		2
#define COLUMN_MTATRANSMITTEDMESSAGES		3
#define COLUMN_MTARECEIVEDVOLUME		4
#define COLUMN_MTASTOREDVOLUME		5
#define COLUMN_MTATRANSMITTEDVOLUME		6
#define COLUMN_MTARECEIVEDRECIPIENTS		7
#define COLUMN_MTASTOREDRECIPIENTS		8
#define COLUMN_MTATRANSMITTEDRECIPIENTS		9
#define COLUMN_MTASUCCESSFULCONVERTEDMESSAGES		10
#define COLUMN_MTAFAILEDCONVERTEDMESSAGES		11
#define COLUMN_MTALOOPSDETECTED		12

/*
 * column number definitions for table mtaGroupErrorTable 
 */
#define COLUMN_MTAGROUPINBOUNDERRORCOUNT		1
#define COLUMN_MTAGROUPINTERNALERRORCOUNT		2
#define COLUMN_MTAGROUPOUTBOUNDERRORCOUNT		3
#define COLUMN_MTASTATUSCODE		4
#endif                          /* MTA_COLUMNS_H */
